unit TransBmp;

interface

uses
	WinTypes, WinProcs, Graphics, Classes, SysUtils;

type
	{Promotes protected TBitmap methods to public access}
	TPublicBitmap = class(TBitmap)
	public
		procedure ReadData(Stream: TStream);
		procedure WriteData(Stream: TStream);
	end;

	TTransparentBitmap = class(TPersistent)
	private
		FAndMask: TBitmap;
		FOrMask: TBitmap;
		FBitmap: TBitmap;
		FTransColor: TColor;
		FTransPixel: TPoint;
		FChanging: Boolean;
		FTransparent: Boolean;
		procedure SetBitmap(Value: TBitmap);
		procedure BitmapChanged(Sender: TObject);
		procedure MakeMasks;
		procedure SetTransparent(Value: Boolean);
		procedure SetTransPixel(Value: TPoint);

	protected
		procedure DefineProperties(Filer: TFiler); override;
		procedure ReadMasks(Stream: TStream);
		procedure WriteMasks(Stream: TStream);

	public
		constructor Create; virtual;
		destructor Destroy; override;
		procedure Assign(Value: TPersistent); override;

		property AndMask: TBitmap read FAndMask;		{Public run-time only to allow access to these bitmaps}
		property OrMask: TBitmap read FOrMask;
		property TransparentPixel: TPoint read FTransPixel write SetTransPixel;

		property Bitmap: TBitmap read FBitmap write SetBitmap;
		property Transparent: Boolean read FTransparent write SetTransparent default True;

{	published
}	end;

implementation

{ TPublicBitmap allows accessing protected methods in TBitmap. Can't use
	the normal LoadFromStream/SaveToStream methods for streaming a control
  to a DFM file. These will generate warnings in Delphi 3.}
procedure TPublicBitmap.ReadData(Stream: TStream);
begin
	inherited ReadData(Stream);
end;

procedure TPublicBitmap.WriteData(Stream: TStream);
begin
	inherited WriteData(Stream);
end;

procedure TTransparentBitmap.DefineProperties(Filer: TFiler);	{Stream masks only; FBitmap streams itself}
begin
	inherited DefineProperties(Filer);
	Filer.DefineBinaryProperty('Masks', ReadMasks, WriteMasks, (FBitmap.Handle <> 0));
end;

procedure TTransparentBitmap.ReadMasks(Stream: TStream);
begin
	TPublicBitmap(FAndMask).ReadData(Stream);
	TPublicBitmap(FOrMask).ReadData(Stream);
end;

procedure TTransparentBitmap.WriteMasks(Stream: TStream);
begin
	TPublicBitmap(FAndMask).WriteData(Stream);
	TPublicBitmap(FOrMask).WriteData(Stream);
end;

constructor TTransparentBitmap.Create;
begin
	inherited Create;
	FTransparent := True;
	FBitmap := TBitmap.Create;
	FBitmap.OnChange := BitmapChanged;
	FAndMask := TBitmap.Create;
	FOrMask := TBitmap.Create;

	FTransPixel.X := 0;
	FTransPixel.Y := 0;
end;

destructor TTransparentBitmap.Destroy;
begin
	FAndMask.Free;
	FOrMask.Free;
	FBitmap.Free;
	inherited Destroy;
end;

procedure TTransparentBitmap.BitmapChanged(Sender: TObject);
begin
	if FChanging then	 {If currently making masks, don't reenter}
		exit;

	FChanging := True;
	if not FTransparent then
	begin								{Keeps resource use to a minimum}
		FAndMask.Free;
		FAndMask := nil;
		FOrMask.Free;
		FOrMask := nil;
		FChanging := False;
	end else
	begin
		try
			if FAndMask = nil then
				FAndMask := TBitmap.Create;
			if FOrMask = nil then
				FOrMask := TBitmap.Create;
			MakeMasks;
		finally
			FChanging := False;
		end;
	end;
end;

procedure TTransparentBitmap.MakeMasks;
var
	TranspColor, SaveBkColor: TColor;
	TempMask: TBitmap;
	SavePal: HPalette;
	DebugDC, DebugOrigPal: THandle;
begin
	if (FBitmap.Width <= 0) or (FBitmap.Height <= 0) then
	begin
		FAndMask.Width := 0;
		FAndMask.Height := 0;
		FOrMask.Width := 0;
		FOrMask.Height := 0;
	end else
	begin
		TempMask := TBitmap.Create;		{Make a monochrome bitmap which holds the AND mask}
		try
			TranspColor := FBitmap.Canvas.Pixels[FTransPixel.X, FTransPixel.Y] + $02000000;

			TempMask.Width := FBitmap.Width;
			TempMask.Height := FBitmap.Height;

			FOrMask.Width := FBitmap.Width;
			FOrMask.Height := FBitmap.Height;

			TempMask.Monochrome := True;
			FAndMask.Width := FBitmap.Width;
			FAndMask.Height := FBitmap.Height;
{
			DebugDC := GetDC(0);
			DebugOrigPal := SelectPalette(DebugDC, FBitmap.Palette, False);
			RealizePalette(DebugDC);
}
			{Set background color of the source to the color in the bitmap that should be transparent}
			SaveBkColor := SetBkColor(FBitmap.Canvas.Handle, TranspColor);
			BitBlt(TempMask.Canvas.Handle, 0, 0, TempMask.Width, TempMask.Height, FBitmap.Canvas.Handle, 0, 0, SRCCOPY);
			SetBkColor(FBitmap.Canvas.Handle, SaveBkColor);	{Reset background color}
			BitBlt(FAndMask.Canvas.Handle, 0, 0, FAndMask.Width, FAndMask.Height, TempMask.Canvas.Handle, 0, 0, SRCCOPY);
{
BitBlt(DebugDC, 0, 0, FAndMask.Width, FAndMask.Height, FAndMask.Canvas.Handle, 0, 0, SRCCOPY);
}
			BitBlt(TempMask.Canvas.Handle, 0, 0, TempMask.Width, TempMask.Height, FAndMask.Canvas.Handle, 0, 0, DSTINVERT);
			BitBlt(FOrMask.Canvas.Handle, 0, 0, FOrMask.Width, FOrMask.Height, TempMask.Canvas.Handle, 0, 0, SRCCOPY);
{
BitBlt(DebugDC, FOrMask.Width, FOrMask.Height, FOrMask.Width, FOrMask.Height, FOrMask.Canvas.Handle, 0, 0, SRCCOPY);
}
			BitBlt(FOrMask.Canvas.Handle, 0, 0, FOrMask.Width, FOrMask.Height, FBitmap.Canvas.Handle, 0, 0, SRCAND);
{
			DebugOrigPal := SelectPalette(DebugDC, DebugOrigPal, False);
			ReleaseDC(0, DebugDC);
}		finally
			TempMask.Free;
		end;
	end;
end;

procedure TTransparentBitmap.SetBitmap(Value: TBitmap);
begin
	if not FBitmap.Empty then
		FBitmap.Free;
	FBitmap := Value;
	BitmapChanged(FBitmap);		{This makes new FAndMask and FOrMask}
end;

procedure TTransparentBitmap.Assign(Value: TPersistent);
begin
	if Value is TBitmap then
	begin
		FBitmap.Assign(Value);
		BitmapChanged(FBitmap);		{This makes new FAndMask and FOrMask}
	end else
  	raise Exception.Create('Assignment requires a TBitmap argument');
end;

procedure TTransparentBitmap.SetTransPixel(Value: TPoint);
begin
	if (Value.X = FTransPixel.X) and (Value.Y = FTransPixel.Y) then
		exit;

	FTransPixel.X := Value.X;
	FTransPixel.Y := Value.Y;

  BitmapChanged(FBitmap);
end;

procedure TTransparentBitmap.SetTransparent(Value: Boolean);
begin
	if Value <> FTransparent then
	begin
		FTransparent := Value;
		BitmapChanged(FBitmap);		{This makes new FAndMask and FOrMask}
	end;
end;

end.

